/* Squint - Pentomino placement game
 * Copyright (C) 2005 David Turner <novalis@novalis.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110
 */

#include "SquintCanvas.h"

#include <iostream>
using namespace std;

SquintCanvas::SquintCanvas (QWidget * parent, const char * name, Qt::WFlags f) :
  QCanvasView (parent, name, f), current_move (0), turn(0), self(0), other(0),
  dragging_item (0)
{
  _canvas = new QCanvas();
  _canvas->resize(600, 300);
  setCanvas(_canvas);

  setHScrollBarMode(QScrollView::AlwaysOff);
  setVScrollBarMode(QScrollView::AlwaysOff);

}

SquintCanvas::~SquintCanvas() {
  if (current_move) delete current_move;
}

static void addPiece (SquintPieceItem *piece, int &x, int &y) {
  piece->setHome(QPoint(x, y));
  x += 100;
  if(x >= 600){
    y += 100;
    if(y < 200)
      x = 300;
    else
      x = 0;
  }
}

void SquintCanvas::cancelMove() {
  delete current_move;
  current_move = 0;
}

void SquintCanvas::flip(int axis) {
  cancelMove();
  QCanvasItemList l = _canvas->allItems();    
  for(QCanvasItemList::iterator i = l.begin(), e = l.end(); i != e; ++i) {
    if ((*i)->rtti() == SquintPieceItem::RTTI) {
      SquintPieceItem *piece = dynamic_cast<SquintPieceItem*>(*i);
      if (axis == 0) 
	piece->doHFlip();
      else
	piece->doVFlip();

      piece->goHome();
    }
  }
  _canvas->update();
}

void SquintCanvas::rotate(int amount) {
  cancelMove();
  if (dragging_item) {
    dragging_item = 0;
  }
  QCanvasItemList l = _canvas->allItems();    
  for(QCanvasItemList::iterator i = l.begin(), e = l.end(); i != e; ++i) {
    if ((*i)->rtti() == SquintPieceItem::RTTI) {
      SquintPieceItem* piece = dynamic_cast<SquintPieceItem*>(*i);

      piece->doRotate(amount);
      piece->goHome();
    }
  }
  _canvas->update();
}


SquintPieceItem* SquintCanvas::findPieceByName(const QString& name) {

  QCanvasItemList l = _canvas->allItems();    
  for(QCanvasItemList::iterator i = l.begin(), e = l.end(); i != e; ++i) {
    if ((*i)->rtti() == SquintPieceItem::RTTI) {
      SquintPieceItem* piece = dynamic_cast<SquintPieceItem*>(*i);
      if (piece->piece.name == name) {
	return piece;
      }
    }
  }
  return 0;
}

void SquintCanvas::buildGame () {
  started = false;
  cancelMove();
  QCanvasItemList l = _canvas->allItems();    
  for(QCanvasItemList::iterator i = l.begin(), e = l.end(); i != e; ++i) {
    delete *i;
  }

  for (int y = 0; y < 9; ++y) {
    for (int x = 0; x < 12; ++x) {
      board[y][x] = 0;
    }
  }

  int x = 300, y = 0;
    
  board_item = new SquintBoard(this);
  dragging_item = 0;

  addPiece (new SquintPieceItem (_canvas, "11111", 5, 1, "I"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "1111"
				 "0001", 4, 2, "L"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "1111"
				 "0010", 4, 2, "Y"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "1110"
				 "0011", 4, 2, "Z"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "111"
				 "011", 3, 2, "P"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "111"
				 "101", 3, 2, "U"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "100"
				 "110"
				 "011", 3, 3, "W"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "011"
				 "110"
				 "010", 3, 3, "F"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "010"
				 "010"
				 "111", 3, 3, "T"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "010"
				 "111"
				 "010", 3, 3, "X"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "100"
				 "111"
				 "001", 3, 3, "N"), x, y);
  addPiece (new SquintPieceItem (_canvas, 
				 "100"
				 "100"
				 "111", 3, 3, "V"), x, y);

  _canvas->update();
}

bool SquintCanvas::goodPlace(SquintPieceItem *pieceItem, int gridx, int gridy) {

  SquintPiece& piece = pieceItem->piece;

  /* if this is the first move, check the corner*/
  if (!started) {
    if (gridx == 0) {
      if (gridy == 0) {
	if (!piece.grid[0][0])
	  return false;
      } else {
	if (gridy + piece.height != 9)
	  return false;
	if (!piece.grid[piece.height - 1][0])
	  return false;
      } 
    } else {
      if (gridx + piece.width != 12)
	return false;
      if (gridy == 0) {
	if (!piece.grid[0][piece.width - 1])
	  return false;
      } else {
	if (gridy + piece.height != 9)
	  return false;
	if (!piece.grid[piece.height - 1][piece.width - 1])
	  return false;
      } 
    }
    return true;
  }


  //check for overlap.  Also, check for nextto-ness
  for (int y = 0; y < piece.height; y++) {
    for (int x = 0; x < piece.width; x++) {
      if (piece.grid[y][x] && board[y+gridy][x+gridx]) 
	return false;
      if (y+gridy+1 < 9) 
	if (piece.grid[y][x] && board[y+gridy+1][x+gridx]) 
	  return false;
      if (y+gridy-1 >= 0) 
	if (piece.grid[y][x] && board[y+gridy-1][x+gridx]) 
	  return false;
      if (x+gridx+1 < 12) 
	if (piece.grid[y][x] && board[y+gridy][x+gridx+1]) 
	  return false;
      if (x+gridx-1 >= 0) 
	if (piece.grid[y][x] && board[y+gridy][x+gridx-1]) 
	  return false;
    }
  }


  //otherwise, we need to check the corners of the piece
  for (int y = 0; y < piece.height; y++) {
    for (int x = 0; x < piece.width; x++) {
      if (y+gridy+1 < 9) {
	if (x+gridx+1 < 12) 
	  if (piece.grid[y][x] && board[y+gridy+1][x+gridx+1]) 
	    return true;
	if (x+gridx-1 > 0) 
	  if (piece.grid[y][x] && board[y+gridy+1][x+gridx-1]) 
	    return true;
      }
      if (y+gridy-1 >= 0) {
	if (x+gridx+1 < 12) 
	  if (piece.grid[y][x] && board[y+gridy-1][x+gridx+1]) 
	    return true;
	if (x+gridx-1 >= 0) 
	  if (piece.grid[y][x] && board[y+gridy-1][x+gridx-1]) 
	    return true;
      }
    }
  }
  return false;

}


void SquintCanvas::takeTurn(SquintMove* move) {

  placePiece(move->piece, move->x, move->y);
  move->piece->hide();
  _canvas->update();
  started = true;
  if (turn == self) {
    turn = other;
  } else {
    turn = self;
  }
}

void SquintCanvas::placePiece(SquintPieceItem* piece, int gridx, int gridy) {

  //hide piece, fill board
  piece->hide();

  for (int y = 0; y < piece->piece.height; y++) {
    for (int x = 0; x < piece->piece.width; x++) {
      board[y+gridy][x+gridx] |= piece->piece.grid[y][x];
    }
  }
  board_item->update();
}

void SquintCanvas::contentsMouseReleaseEvent(QMouseEvent *) {

  if (!dragging_item)
    return;

  QPoint corner = dragging_item->corner();

  int gridx = (corner.x() + SQUARE_WIDTH / 2) / SQUARE_WIDTH;
  int gridy = (corner.y() + SQUARE_WIDTH / 2) / SQUARE_WIDTH;
  if (gridx < 0 || gridx + dragging_item->piece.width > 12 || 
      gridy < 0 || gridy + dragging_item->piece.height > 9) {
    dragging_item->goHome();
    dragging_item = 0;
    _canvas->update();
    return;
  }

  //check this location
  if (!goodPlace(dragging_item, gridx, gridy)) {
    dragging_item->goHome();
    dragging_item = 0;
    _canvas->update();
    return;
  }

  dragging_item->snapToGrid(gridx, gridy);
  if (!current_move) 
    current_move = new SquintMove();
  current_move->player = turn;
  current_move->piece = dragging_item;
  current_move->x = gridx;
  current_move->y = gridy;
  current_move->rotation = dragging_item->piece.rotation;
  current_move->flip = dragging_item->piece.flip;
  
  _canvas->update();

}

void SquintCanvas::contentsMousePressEvent(QMouseEvent *e) {
  if (turn != self) return;
  QPoint point = e->pos();
  QCanvasItemList hits = _canvas->collisions(point);

  for (QCanvasItemList::Iterator it = hits.begin(); it != hits.end(); ++it) {
    if ((*it)->rtti() == SquintPieceItem::RTTI) {

      SquintPieceItem *new_item = dynamic_cast<SquintPieceItem*>(*it);
      if (dragging_item != new_item) {
	if (dragging_item)
	  dragging_item->goHome();
	dragging_item = new_item;
      }

      drag_start = point;
      return;
    }
  }
  dragging_item = 0;
}

void SquintCanvas::contentsMouseMoveEvent(QMouseEvent *e) {
  if (dragging_item) {
    QPoint point = e->pos();
    dragging_item->moveBy(point.x() - drag_start.x(),
			  point.y() - drag_start.y());
    drag_start = point;

    _canvas->update();

  }
}      
        
