/* Squint - Pentomino placement game
 * Copyright (C) 2005 David Turner <novalis@novalis.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110
 */

#include "SquintCanvas.h"
#include "Connect.h"
#include "SquintImpl.h"

#include <iostream>
using namespace std;

#include <qlineedit.h>
#include <qtextbrowser.h>
#include <qstringlist.h>
#include <qregexp.h>
#include <qlabel.h>
#include <qmessagebox.h>

QString& escape (QString& str) {
  return str.replace(QChar(' '), "\x09", false);
}

QString& unescape (QString& str) {
  return str.replace(QChar(9), " ", false);
}


SquintImpl::SquintImpl () :
  connecting(0), socket(0), games(0)
{
}

void SquintImpl::hFlip() {
  canvas->flip(0);
}

void SquintImpl::vFlip() {
  canvas->flip(1);
}

void SquintImpl::rotateCW() {
  canvas->rotate(1);
}

void SquintImpl::rotateCCW() {
  canvas->rotate(3);
}

void SquintImpl::fileNew() {
  Connect _connect(this);
  if (!_connect.exec()) return;

  /*do connection*/
  if (socket) {
    delete socket;
  } 
  name = _connect.username->text();
  QString server = _connect.server->text();

  connecting = new ConnectingImpl(this);
  connecting->server->setText(server);
  connecting->setModal(true);
  connect (connecting, SIGNAL(cancelled()), this, SLOT(cancelled()));

  connecting->show();

  socket = new QSocket();
  connect(socket, SIGNAL(connected()), this, SLOT(connected()));
  socket->connectToHost(server, 10648);
}

void SquintImpl::move(const QStringList& args) {
  SquintPieceItem* piece = canvas->findPieceByName(args[1]);
  SquintMove move = SquintMove (canvas->turn, piece, args);

  if (canvas->turn == canvas->other)
    showChat("Server", QString("Opponent played ") + piece->piece.name);

  canvas->takeTurn(&move);
}

void SquintImpl::cancelled() {
  if (socket) {
    socket->close();
    delete socket;
    socket = 0;
  }
  delete connecting;
  connecting = 0;
}

void SquintImpl::connected() {

  delete connecting;
  connecting = 0;

  send("login " + name);

  connect(socket, SIGNAL(readyRead()), this, SLOT(read()));
  connect(socket, SIGNAL(delayedCloseFinished()), this, SLOT(disconnected()));
  connect(socket, SIGNAL(connectionClosed()), this, SLOT(disconnected()));

  games = new GamesImpl (this);
  while (1) {
    if (games->exec()) {
      if (games->game) {
	//joining an existing game
	send("join " + escape(games->game));
	break;
      } else {
	QString name = games->newGame->text();
	if (!name) continue;
	send("create " + escape(name));
	showChat("Server", "Created a new game, waiting for someone to join.");
	break;
      }
      canvas->self = new SquintPlayer (name);
    } else {
      socket->close();
      delete socket;
      break;
    }
  }
  delete games;
  games = 0;
}

void SquintImpl::read() {
  while (socket->canReadLine()) {
    QString line = socket->readLine();
    QStringList tokens = QStringList::split(QRegExp("[ \r\n][ \r\n]*"), line);

    //cout << line << endl;

    if (tokens[0] == "starting") {
      if (tokens[1] == "you") {
	showChat ("Server", "Game started.  Your turn.");
	canvas->turn = canvas->self;
      } else {
	showChat ("Server", "Game started.  Opponent's turn.");
	canvas->turn = canvas->other;
      }
      canvas->buildGame();
    } else if (tokens [0] == "joined") {
      canvas->other = new SquintPlayer (tokens[1]);
      showChat (canvas->other->name, "/me joined");
    } else if (tokens [0] == "youjoined") {
      canvas->self = new SquintPlayer (name);
    } else if (tokens [0] == "chat") {
      if (canvas->other) {
	showChat (canvas->other->name, tokens[1]);
      } else {
	showChat ("Server", tokens[1]);
      }
    } else if (tokens [0] == "move") {
      move(tokens);
    } else if (tokens [0] == "announce") {
      if (games) {
	games->addGame(tokens[1], unescape (tokens[2]));
      }
    } else if (tokens [0] == "unannounce") {
      if (games) {
	games->removeGame(unescape (tokens[1]));
      }
    }
  }
}

void SquintImpl::showChat(const QString& who, const QString& what) {
  QString chat;
  QString whatclean = what;
  unescape (whatclean);

  if (whatclean.startsWith("/me ")) {
    chat += "*";
    chat += who;
    chat += " ";
    chat += whatclean.mid(4);
  } else {
    chat += "<";
    chat += who;
    chat += "> ";
    chat += whatclean;
  }

  chat += "\n";
  chatLog->append (chat);
}

void SquintImpl::sendMove() {
  if (canvas->current_move) {
    send("move " + canvas->current_move->serialize());
    canvas->cancelMove();
  }
}

void SquintImpl::send(const QString& data) {
  //cout << "SENDING: " << data << endl;
  QString tdata = data + "\n";
  socket->writeBlock(tdata.utf8(), tdata.length());
}

void SquintImpl::send(const QStringList& data) {
  QString tdata;
  for (QStringList::ConstIterator it = data.begin(); it != data.end(); ++it ) {
    QString str (*it);
    tdata += str.replace(QChar(' '), "\x09");
    tdata += " ";
  }
  send(tdata);
  socket->flush();
}

void SquintImpl::disconnected() {

  //we need to notify the user of the problem and let them log back on.

}


void SquintImpl::closeEvent(QCloseEvent* ce) {

  int result = QMessageBox::information (this, "Squint",
					 "Are you sure you want to quit?",
					 tr("&Don't Quit"), tr("&Quit"));
  if (result == 0)
        ce->ignore();
  else {
        ce->accept();
  }
}

void SquintImpl::sendChat() {
  QString text = chatInput->text();
  chatInput->clear();

  QStringList args;
  args << "chat";
  args << text;
  send(args);
  showChat(name, text);
}

void SquintImpl::restartGame() {

  if (!socket) {
    fileNew();
    return;
  }

  int result = QMessageBox::information (this, "Squint",
					 "Are you sure you want to restart?",
					 tr("&Don't Restart"), tr("&Restart"));
  if (result == 1) {
    send(QString("restart"));
  }
}
