/* Squint - Pentomino placement game
 * Copyright (C) 2005 David Turner <novalis@novalis.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110
 */

#include <qstringlist.h>
#include <qregexp.h>

#include "SquintClient.h"

#include <iostream>
using namespace std;

SquintClient::SquintClient(QSocket* socket, SquintServer* server) :
QObject(server), game(0), dead(false), socket(socket), server(server)
{

  connect(socket, SIGNAL(readyRead()), this, SLOT(read()));
  connect(socket, SIGNAL(connectionClosed()), this, SLOT(disconnected()));
}

SquintClient::~SquintClient() {
  if (socket) {
    delete socket;
    socket = 0;
  }
  if (game) {
    delete game;
    game = 0;
  }
}

  /* fixme: next four are also in SquintImpl */

QString& escape (QString& str) {
  return str.replace(QChar(' '), "\x09", false);
}

QString& unescape (QString& str) {
  return str.replace(QChar(9), " ", false);
}

void SquintClient::send(const QString& data) {
  if (dead) return; //Son, you can talk to the dead, but they won't talk back
  //cout << "SENDING: " << data << endl;
  QString tdata = data + "\n";
  socket->writeBlock(tdata.utf8(), tdata.length());
  socket->flush();
}

void SquintClient::send(const QStringList& data) {
  QString tdata;
  for (QStringList::ConstIterator it = data.begin(); it != data.end(); ++it ) {
    QString str (*it);
    tdata += str.replace(QChar(' '), "\x09");
    tdata += " ";
  }
  send(tdata);
}

void SquintClient::createGame(const QString& gameName) {
  if (!name) {
    return; //fixme: error correction
  }
  if (!game) {
    game = new SquintGame(gameName, this);
    server->newGame(game);
  } else {
    game->restart(gameName, this);
  }
  announceYouJoin();
}

void SquintClient::announceYouJoin() {
  QStringList args;
  args << QString("youjoined");
  args << name;
  send (args);
}

void SquintClient::announceJoin(SquintClient* player) {
  QStringList args;
  args << QString("joined");
  args << player->name;
  send (args);
}

void SquintClient::announceGame(SquintGame* game) {
  QStringList args;
  args << QString("announce");
  args << game->players[0]->name;
  args << game->name;
  send (args);
}

void SquintClient::unannounceGame(SquintGame* game) {
  QStringList args;
  args << QString("unannounce");
  args << game->name;
  send (args);
}

void SquintClient::sendMove(const QString& move) {
  send (move);
}


void SquintClient::read() {
  if (!socket->canReadLine()) 
    return;

  QString line = socket->readLine();
  QStringList tokens = QStringList::split(QRegExp("[ \r\n][ \r\n]*"), line);

  //cout << "READING: " << line << endl;

  if (tokens[0] == "login") {
    name = tokens[1];
    server->sendGames(this);
  } else if (tokens[0] == "chat") {
    game->chat(tokens[1], this);
  } else if (tokens[0] == "move") {
    if (game->isTurn(this))
	game->move(line);
    else
      sendError("It's not your turn");
  } else if (tokens [0] == "create") {
    unescape(tokens[1]);
    createGame(tokens[1]);
  } else if (tokens [0] == "restart") {
    game->startGame();
  } else if (tokens [0] == "join") {
    if (!name) {
      return; //fixme: error correction
    }
    if (game) {
      return; //fixme: error correction
    }
    game = server->findGameByName(tokens[1]);
    if (!game || !server->removeGame(game)) {
      sendError ("No such game");
      return;
    }
    game->join(this);
  }

}

void SquintClient::sendError (const QString& error) {
  send("error " + error);
}

void SquintClient::startGame (bool yourTurn) {
  QStringList args;
  args << QString("starting");
  args << (yourTurn ? QString("you") : QString("them"));
  send(args);
}

void SquintClient::disconnected() {
  if (game)
    game->clientDisconnected(this);
  server->discardClient(this);
  delete socket;
  socket = 0;
  dead = true;
}

void SquintClient::partnerDisconnected() {
  send (QString("partnerdisconnected"));
}
